/*
 * $Header: //depot/php_ext/cTemplate/php_cTemplate.cpp#12 $
 */

/*!
 * \file php_cTemplate.cpp
 * \brief that's php extension for google ctemplate.\n
 *        here is example.\n
 *
 *        $tpl = new ctemplate ("example.tpl", STRIP_BLANK_LINES);\n
 *        $dict = new ctemplate_dict;\n
 *        $dict->set ("NAME", "John Smith");\n
 *        $out = $tpl->expand ($dict);\n
 *        echo $out;\n
 */
/* {{{ include files */
extern "C" {
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "ext/standard/php_string.h"
#include "php_cTemplate.h"
}

#include "cTemplate.h"
#include "cTemplateDict.h"
/* }}} */

/* {{{ static variable */
static int le_cTemplate;
static int le_cTemplateDict;
static zend_class_entry *ctemplate_class_entry_ptr;
static zend_class_entry *ctemplate_dict_class_entry_ptr;
/* }}} */

/* {{{ function_entry */
zend_function_entry cTemplate_functions[] = {
    PHP_FE (ctemplate_reload, NULL)
    PHP_FE (ctemplate_clearcache, NULL)
    {NULL, NULL, NULL}
};

static zend_function_entry ctemplate_class_functions[] = {
    PHP_FE (ctemplate, NULL)
    PHP_FALIAS (expand, t_Expand, NULL)
    PHP_FALIAS (dump, t_Dump, NULL)
    PHP_FALIAS (state, t_state, NULL)
    PHP_FALIAS (template_file, t_template_file, NULL)
    PHP_FALIAS (reloadifchanged, t_ReloadIfChanged, NULL)
    PHP_FALIAS (writeheaderentries, t_WriteHeaderEntries, NULL)
    {NULL, NULL, NULL}
};

static zend_function_entry ctemplate_dict_class_functions[] = {
    PHP_FE (ctemplate_dict, NULL)
    PHP_FALIAS (set, d_SetValue, NULL)
    PHP_FALIAS (setescaped, d_SetEscapedValue, NULL)
    PHP_FALIAS (setglobal, d_SetGlobalValue, NULL)
    PHP_FALIAS (addsection, d_AddSectionDictionary, NULL)
    PHP_FALIAS (show, d_ShowSection, NULL)
    PHP_FALIAS (addinclude, d_AddIncludeDictionary, NULL)
    PHP_FALIAS (filename, d_SetFilename, NULL)
    PHP_FALIAS (dump, d_Dump, NULL)
    PHP_FALIAS (dumptostring, d_DumpToString, NULL)
    PHP_FALIAS (setannotateoutput, d_SetAnnotateOutput, NULL)
    {NULL, NULL, NULL}
};
/* }}} */

/* {{{ module entry */
zend_module_entry cTemplate_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    "cTemplate",
    cTemplate_functions,
    PHP_MINIT(cTemplate),
    PHP_MSHUTDOWN(cTemplate),
    NULL,
    NULL,
    PHP_MINFO(cTemplate),
#if ZEND_MODULE_API_NO >= 20010901
    "$Revision: #12 $",
#endif
    STANDARD_MODULE_PROPERTIES
};
/* }}} */

/* {{{ resource allocate function */
static void * _return_resource (zval *wrapper, int rsrc_type TSRMLS_DC)
{
    void *obj = NULL;
    zval **handle;
    int type;

    if (Z_TYPE_P(wrapper) != IS_OBJECT)
    {
        php_error (E_ERROR, "Wrapper is not an object");
    }

    if (zend_hash_find (Z_OBJPROP_P(wrapper), "id", sizeof ("id"), (void **) &handle) == FAILURE)
    {
        php_error (E_ERROR, "Underlying object missing1");
    }

    obj = zend_list_find (Z_LVAL_PP (handle), &type);
    if (!obj || (type != rsrc_type))
    {
        php_error (E_ERROR, "Underlying object missing2");
    }
    return obj;
}

static void _free_ctemplate_resource (zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    cTemplate *obj = static_cast <cTemplate *> (rsrc->ptr);

    if (obj)
    {
        delete obj;
        rsrc->ptr = NULL;
    }
}

static void _free_ctemplate_dict_resource (zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    cTemplateDict *obj = static_cast <cTemplateDict *> (rsrc->ptr);

    if (obj)
    {
        delete obj;
        rsrc->ptr = NULL;
    }
}

static void _add_resource (zval *wrapper, void *obj, int rsrc_type TSRMLS_DC)
{
    zval *handle = NULL;

    if (rsrc_type == le_cTemplate)
        object_init_ex (wrapper, ctemplate_class_entry_ptr);
    else if (rsrc_type == le_cTemplateDict)
        object_init_ex (wrapper, ctemplate_dict_class_entry_ptr);

    MAKE_STD_ZVAL (handle);
    ZEND_REGISTER_RESOURCE (handle, obj, rsrc_type);
    zend_hash_update (Z_OBJPROP_P (wrapper), "id", sizeof ("id"), &handle, sizeof (zval *), NULL);
}

/* }}} */

/* {{{ INIT */
extern "C" {
#ifdef COMPILE_DL_CTEMPLATE
ZEND_GET_MODULE(cTemplate)
#endif
}

PHP_MINIT_FUNCTION(cTemplate)
{
    zend_class_entry ctemplate_ce;
    zend_class_entry ctemplate_dict_ce;
    
    le_cTemplate = zend_register_list_destructors_ex (_free_ctemplate_resource, NULL, "ctemplate", module_number);
    le_cTemplateDict = zend_register_list_destructors_ex (_free_ctemplate_dict_resource, NULL, "ctemplate_dict", module_number);

    INIT_CLASS_ENTRY (ctemplate_ce, "ctemplate", ctemplate_class_functions);
    INIT_CLASS_ENTRY (ctemplate_dict_ce, "ctemplate_dict", ctemplate_dict_class_functions);

    ctemplate_class_entry_ptr = zend_register_internal_class (&ctemplate_ce TSRMLS_CC);
    ctemplate_dict_class_entry_ptr = zend_register_internal_class (&ctemplate_dict_ce TSRMLS_CC);

    REGISTER_LONG_CONSTANT ("DO_NOT_STRIP", DO_NOT_STRIP, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("STRIP_BLANK_LINES", STRIP_BLANK_LINES, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("STRIP_WHITESPACE", STRIP_WHITESPACE, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("NUM_STRIPS", NUM_STRIPS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("TS_UNUSED", TS_UNUSED, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("TS_EMPTY", TS_EMPTY, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("TS_ERROR", TS_ERROR, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("TS_READY", TS_READY, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("TS_RELOAD", TS_RELOAD, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("HTML_ESCAPE", 0, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("XML_ESCAPE", 1, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT ("JAVASCRIPT_ESCAPE", 2, CONST_CS | CONST_PERSISTENT);
    return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(cTemplate)
{
    return SUCCESS;
}

PHP_MINFO_FUNCTION(cTemplate)
{
    php_info_print_table_start();
    php_info_print_table_header(2, "cTemplate support", "enabled");
    php_info_print_table_end();
}
/* }}} */

/* {{{ ctemplate */
/*
 * $tpl = new ctempate (string $filename, int $strip_option, [string $rootdir]);
 */
PHP_FUNCTION(ctemplate)
{
    cTemplate * ctemplate;
    Template *tpl = NULL;
    const char *fn = NULL, *dir = NULL;
    int fn_len, dir_len;
    long flags;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "sl|s", &fn, &fn_len, &flags, &dir, &dir_len) == FAILURE)
        RETURN_FALSE;

    if (dir != NULL)
        Template::SetTemplateRootDirectory (dir);
    else
        Template::SetTemplateRootDirectory ("./");

    tpl = Template::GetTemplate (fn, (Strip)flags);

    if (tpl == NULL)
    {
        php_error (E_ERROR, "Coult not create ctemplate object");
        RETURN_FALSE;
    }

    ctemplate = new cTemplate;
    ctemplate->t = tpl;
    _add_resource (getThis(), ctemplate, le_cTemplate TSRMLS_CC);
    RETURN_TRUE;
}
/* }}} */

/* {{{ ctemplate_reload */
/*
 * bool = ctemplate_reload ();
 */
PHP_FUNCTION(ctemplate_reload)
{
    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    Template::ReloadAllIfChanged();
    RETURN_TRUE;
}
/* }}} */

/* {{{ ctemplate_clearcache */
/*
 * bool = ctemplate_clearcache();
 */
PHP_FUNCTION(ctemplate_clearcache)
{
    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    Template::ClearCache ();
    RETURN_TRUE;
}
/* }}} */

/* {{{ t_Expand */
/*
 * string $out = $t->expand (dict_object $dict);
 */
PHP_FUNCTION(t_Expand)
{
    zval *d;
    cTemplate *ctemplate = NULL;
    cTemplateDict *cd = NULL;
    string ret;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "o", &d) == FAILURE)
        RETURN_FALSE;

    ctemplate = (cTemplate *)_return_resource (getThis(), le_cTemplate TSRMLS_CC);
    if (ctemplate == NULL)
    {
        php_error (E_WARNING, "t_Expand: error getting object");
        RETURN_FALSE;
    }

    cd = (cTemplateDict *)_return_resource (d, le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "t_Expand: error getting dict object");
        RETURN_FALSE;
    }

    if (cd->root == true)
        ctemplate->t->Expand (&ret, &(cd->d));
    else
        ctemplate->t->Expand (&ret, cd->p);

    RETURN_STRINGL ((char *)ret.c_str(), ret.length(), 1);
}
/* }}} */

/* {{{ t_Dump */
/*
 * bool = $t->dump (string $fn);
 */
PHP_FUNCTION(t_Dump)
{
    const char *fn = NULL;
    int fn_len;
    cTemplate *ctemplate = NULL;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "s", &fn, &fn_len) == FAILURE)
        RETURN_FALSE;

    ctemplate = (cTemplate *)_return_resource (getThis(), le_cTemplate TSRMLS_CC);
    if (ctemplate == NULL)
    {
        php_error (E_WARNING, "t_Dump: error getting object");
        RETURN_FALSE;
    }

    ctemplate->t->Dump (fn);
    RETURN_TRUE;
}
/* }}} */

/* {{{ t_state */
/*
 * $state = $tpl->state();
 */
PHP_FUNCTION(t_state)
{
    cTemplate *ctemplate = NULL;

    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    ctemplate = (cTemplate *)_return_resource (getThis(), le_cTemplate TSRMLS_CC);
    if (ctemplate == NULL)
    {
        php_error (E_WARNING, "t_state: error getting object");
        RETURN_FALSE;
    }

    RETURN_LONG (ctemplate->t->state());
}
/* }}} */

/* {{{ t_template_file */
/*
 * string $fn = $tpl->template_file ();
 */
PHP_FUNCTION(t_template_file)
{
    cTemplate *ctemplate = NULL;

    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    ctemplate = (cTemplate *)_return_resource (getThis(), le_cTemplate TSRMLS_CC);
    if (ctemplate == NULL)
    {
        php_error (E_WARNING, "t_template_file: error getting object");
        RETURN_FALSE;
    }

    RETURN_STRING ((char *)ctemplate->t->template_file(), 1);
}
/* }}} */

/* {{{ t_ReloadIfChanged */
/*
 * bool = $tpl->reloadifchanged ();
 */
PHP_FUNCTION(t_ReloadIfChanged)
{
    cTemplate *ctemplate = NULL;
    zend_bool b;

    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    ctemplate = (cTemplate *)_return_resource (getThis(), le_cTemplate TSRMLS_CC);
    if (ctemplate == NULL)
    {
        php_error (E_WARNING, "t_ReloadIfChanged: error getting object");
        RETURN_FALSE;
    }

    b = ctemplate->t->ReloadIfChanged();
    RETURN_BOOL (b);
}
/* }}} */

/* {{{ t_WriteHeaderEntries */
/*
 * string $header = $tpl->writeheaderentries ();
 */
PHP_FUNCTION(t_WriteHeaderEntries)
{
    cTemplate *ctemplate = NULL;
    string ret;

    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    ctemplate = (cTemplate *)_return_resource (getThis(), le_cTemplate TSRMLS_CC);
    if (ctemplate == NULL)
    {
        php_error (E_WARNING, "t_WriteHeaderEntries: error getting object");
        RETURN_FALSE;
    }

    ctemplate->t->WriteHeaderEntries (&ret);
    RETURN_STRINGL ((char *)ret.c_str(), ret.length(), 1);
}
/* }}} */

/* {{{ ctemplate_dict */
/*
 * $dict = new ctemplate_dict ();
 */
PHP_FUNCTION(ctemplate_dict)
{
    cTemplateDict * cd;

    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    cd = new cTemplateDict;

    if (cd != NULL)
    {
        _add_resource (getThis(), cd, le_cTemplateDict TSRMLS_CC);
        RETURN_TRUE;
    }
    RETURN_FALSE;
}
/* }}} */

/* {{{ d_SetValue */
/*
 * bool = $dict->setvalue (string $key, string $val, [string $section]);
 * bool = $dict->setvalue (string $key, int $val, [string $section]);
 */
PHP_FUNCTION(d_SetValue)
{
    cTemplateDict *cd = NULL;
    const char *key = NULL, *val = NULL, *sec = NULL;
    int key_len, val_len, sec_len;
    long lval;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_SetValue: error getting object");
        RETURN_FALSE;
    }

    if (zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET, ZEND_NUM_ARGS() TSRMLS_CC, "sl|s", &key, &key_len, &lval, &sec, &sec_len) == SUCCESS)
    {
        if (sec != NULL)
        {
            if (cd->root == true)
            {
                cd->d.SetIntValue (key, lval);
                cd->d.ShowSection (sec);
            }
            else
            {
                cd->p->SetIntValue (key, lval);
                cd->p->ShowSection (sec);
            }
        }
        else
        {
            if (cd->root == true)
                cd->d.SetIntValue (key, lval);
            else
                cd->p->SetIntValue (key, lval);
        }
        RETURN_TRUE;
    }
    else if (zend_parse_parameters_ex(ZEND_PARSE_PARAMS_QUIET, ZEND_NUM_ARGS() TSRMLS_CC, "ss|s", &key, &key_len, &val, &val_len, &sec, &sec_len) == SUCCESS)
    {
        if (sec != NULL)
        {
            if (cd->root == true)
                cd->d.SetValueAndShowSection (key, val, sec);
            else
                cd->p->SetValueAndShowSection (key, val, sec);
        }
        else
        {
            if (cd->root == true)
                cd->d.SetValue (key, val);
            else
                cd->p->SetValue (key, val);
        }
        RETURN_TRUE;
    }
    else
        RETURN_FALSE;
}
/* }}} */

/* {{{ d_SetEscapedValue */
/*
 * bool = $dict->setescapedvalue (string $key, string $val, long $escape_option);
 */
PHP_FUNCTION(d_SetEscapedValue)
{
    cTemplateDict *cd = NULL;
    const char *key = NULL, *val = NULL, *sec = NULL;
    int key_len, val_len, sec_len;
    long e;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_SetEscapedValue: error getting object");
        RETURN_FALSE;
    }

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ssl|s", &key, &key_len, &val, &val_len, &e, &sec, &sec_len) == FAILURE)
        RETURN_FALSE;

    if (sec != NULL)
    {
        switch (e)
        {
            case 1:
                if (cd->root == true)
                    cd->d.SetEscapedValueAndShowSection (key, val, TemplateDictionary::xml_escape, sec);
                else
                    cd->p->SetEscapedValueAndShowSection (key, val, TemplateDictionary::xml_escape, sec);
                break;
            case 2:
                if (cd->root == true)
                    cd->d.SetEscapedValueAndShowSection (key, val, TemplateDictionary::javascript_escape, sec);
                else
                    cd->p->SetEscapedValueAndShowSection (key, val, TemplateDictionary::javascript_escape, sec);
                break;
            case 0:
            default:
                if (cd->root == true)
                    cd->d.SetEscapedValueAndShowSection (key, val, TemplateDictionary::html_escape, sec);
                else
                    cd->p->SetEscapedValueAndShowSection (key, val, TemplateDictionary::html_escape, sec);
        }
    }
    else
    {
        switch (e)
        {
            case 1:
                if (cd->root == true)
                    cd->d.SetEscapedValue (key, val, TemplateDictionary::xml_escape);
                else
                    cd->p->SetEscapedValue (key, val, TemplateDictionary::xml_escape);
                break;
            case 2:
                if (cd->root == true)
                    cd->d.SetEscapedValue (key, val, TemplateDictionary::javascript_escape);
                else
                    cd->p->SetEscapedValue (key, val, TemplateDictionary::javascript_escape);
                break;
            case 0:
            default:
                if (cd->root == true)
                    cd->d.SetEscapedValue (key, val, TemplateDictionary::html_escape);
                else
                    cd->p->SetEscapedValue (key, val, TemplateDictionary::html_escape);
        }
    }
    RETURN_TRUE;
}
/* }}} */

/* {{{ d_SetGlobalValue */
/*
 * bool = $dict->setglobalvalue (string $key, string $val);
 */
PHP_FUNCTION(d_SetGlobalValue)
{
    cTemplateDict *cd = NULL;
    const char *key = NULL, *val = NULL;
    int key_len, val_len;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "ss", &key, &key_len, &val, &val_len) == FAILURE)
        RETURN_FALSE;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_SetGlobalValue: error getting object");
        RETURN_FALSE;
    }

    if (cd->root == true)
    cd->d.SetGlobalValue (key, val);
    cd->p->SetGlobalValue (key, val);
    RETURN_TRUE;
}
/* }}} */

/* {{{ d_AddSectionDictionary */
/* dict_object $o = $dict->addsectiondictionary (string $another_section);
 */
PHP_FUNCTION(d_AddSectionDictionary)
{
    zval *rv;
    cTemplateDict *cd, *ce;
    const char *sec = NULL;
    int sec_len;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "s", &sec, &sec_len) == FAILURE)
        RETURN_FALSE;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_AddSectionDictionary: error getting object");
        RETURN_FALSE;
    }

    ce = new cTemplateDict;
    ce->p = cd->d.AddSectionDictionary(sec);
    ce->root = false;

    if (ce != NULL)
    {
        MAKE_STD_ZVAL (rv);
        _add_resource (rv, ce, le_cTemplateDict TSRMLS_CC);
        SEPARATE_ZVAL (&rv);
        *return_value = *rv;
        FREE_ZVAL (rv);
        return;
    }
    RETURN_FALSE;
}
/* }}} */

/* {{{ d_ShowSection */
/*
 * bool = $dict->showsection (string $section);
 */
PHP_FUNCTION(d_ShowSection)
{
    cTemplateDict *cd = NULL;
    const char *sec = NULL;
    int sec_len;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "s", &sec, &sec_len) == FAILURE)
        RETURN_FALSE;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_ShowSection: error getting object");
        RETURN_FALSE;
    }

    if (cd->root == true)
        cd->d.ShowSection (sec);
    else
        cd->p->ShowSection (sec);
    RETURN_TRUE;
}
/* }}} */

/* {{{ d_AddIncludeDictionary */
/*
 * dict_object $o = $dict->addincludedictionary (string $another_section);
 */
PHP_FUNCTION(d_AddIncludeDictionary)
{
    zval *rv;
    cTemplateDict *cd, *ce;
    const char *sec = NULL;
    int sec_len;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "s", &sec, &sec_len) == FAILURE)
        RETURN_FALSE;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_AddIncludeDictionary: error getting object");
        RETURN_FALSE;
    }

    ce = new cTemplateDict;
    ce->p = cd->d.AddIncludeDictionary (sec);
    ce->root = false;

    if (ce != NULL)
    {
        MAKE_STD_ZVAL (rv);
        _add_resource (rv, ce, le_cTemplateDict TSRMLS_CC);
        SEPARATE_ZVAL (&rv);
        *return_value = *rv;
        FREE_ZVAL (rv);
        return;
    }
    RETURN_FALSE;
}
/* }}} */

/* {{{ d_SetFilename */
/*
 * bool = $dict->setfilename (string $fn);
 */
PHP_FUNCTION(d_SetFilename)
{
    cTemplateDict *cd = NULL;
    const char *fn = NULL;
    char *out = NULL;
    int fn_len;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "s", &fn, &fn_len) == FAILURE)
        RETURN_FALSE;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_SetFilename: error getting object");
        RETURN_FALSE;
    }

    out = php_trim ((char *)fn, fn_len, NULL, 0, NULL, 3 TSRMLS_CC);

    if (cd->root == true)
        cd->d.SetFilename (out);
    else
        cd->p->SetFilename (out);

    efree (out);
    RETURN_TRUE;
}
/* }}} */

/* {{{ d_Dump */
/*
 * bool = $dict->dump ();
 */
PHP_FUNCTION(d_Dump)
{
    cTemplateDict * cd = NULL;

    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_Dump: error getting object");
        RETURN_FALSE;
    }

    if (cd->root == true)
        cd->d.Dump ();
    else
        cd->p->Dump ();
    RETURN_TRUE;
}
/* }}} */

/* {{{ d_DumpToString */
/*
 * string $out = $dict->dumptostring ();
 */
PHP_FUNCTION(d_DumpToString)
{
    cTemplateDict * cd = NULL;
    string ret;

    if (ZEND_NUM_ARGS() != 0)
    {
        WRONG_PARAM_COUNT;
    }

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_DumpToString: error getting object");
        RETURN_FALSE;
    }

    if (cd->root == true)
        cd->d.DumpToString (&ret);
    else
        cd->p->DumpToString (&ret);
    RETURN_STRINGL ((char *)ret.c_str(), ret.length(), 1);
}
/* }}} */

/* {{{ d_SetAnnotateOutput */
PHP_FUNCTION(d_SetAnnotateOutput)
{
    cTemplateDict * cd = NULL;
    const char *path = NULL;
    int path_len;

    if (zend_parse_parameters( ZEND_NUM_ARGS() TSRMLS_CC, "s", &path, &path_len) == FAILURE)
        RETURN_FALSE;

    cd = (cTemplateDict *)_return_resource (getThis(), le_cTemplateDict TSRMLS_CC);
    if (cd == NULL)
    {
        php_error (E_WARNING, "d_SetAnnotateOutput: error getting object");
        RETURN_FALSE;
    }

    if (cd->root == true)
        cd->d.SetAnnotateOutput (path);
    else
        cd->p->SetAnnotateOutput (path);
    RETURN_TRUE;
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
